import time
from rich.table import Table
import pandas as pd

from myterial import orange, amber


def timestamp(just_time=False):
    """
    Returns a formatted timestamp
    """
    if not just_time:
        return time.strftime("%y%m%d_%H%M%S")
    else:
        return time.strftime("%H:%M:%S")


def _make_table(left_header, right_header, nodim=False):
    tb = Table(box=None)
    tb.add_column(
        left_header,
        header_style=f"{amber}" if nodim else f"{amber} dim",
        style=None if nodim else "dim",
    )
    tb.add_column(right_header, header_style=orange)

    return tb


def as_pandas(data):
    """Returns a dataframe if possible, an error otherwise"""
    if isinstance(data, pd.DataFrame):
        return data
    elif isinstance(data, dict):
        return pd.DataFrame(data)
    else:
        raise TypeError(
            f"Expected a DataFrame or dict type, got: {type(data)} insead"
        )



from pathlib import Path
from fcutils.path import files
import matplotlib.pyplot as plt
import numpy as np

from tpd import recorder

recorder.start()


def test_copy_file():
    recorder.copy(Path(__file__))
    recorder.copy(__file__)

    if not Path(__file__).name in [f.name for f in files(recorder.folder)]:
        raise ValueError("File not copied!")


def test_save_figure():
    f, ax = plt.subplots()

    recorder.add_figure(f, "fig1")
    recorder.add_figure(ax, "f1 from ax", svg=False)

    assert recorder.n_figures == 2

    recorder.describe()
    del f


def test_save_figures():
    f = plt.figure()
    f2 = plt.figure()
    plt.figure()

    f._save_name = "ff1"
    f2._save_name = "ff2"

    recorder.add_figures()

    assert recorder.n_figures == 4


def test_save_numpy():
    X = np.random.rand(100, 100)

    recorder.add_data(
        X, "data", fmt="npy", description="This is my random data"
    )

    assert recorder.n_data == 1


def test_save_matlab():
    X = np.random.rand(100, 100)

    recorder.add_data(
        X, "matlab_data", fmt="mat", description="This is my random data"
    )

    assert recorder.n_data == 2



from tpd import recorder

recorder.start(
    # base_folder='my_computer/my_data',
    # name='simulation_2'
)
"""
    Creates a folder 'simulation_2' in the 'base_folder'
    From now on, all data, logs and plots will be saved in
        my_computer/my_data/simulation_2
"""

# all logs will be saved in a .log file
from loguru import logger

logger.info("I cant wait to check the logs later!")


# you can add data easily
import numpy as np

X = np.random.rand(100, 100)
recorder.add_data(X, "my_results", fmt="npy")  # or 'mat' for matlab


# and save plots
import matplotlib.pyplot as plt

f = plt.figure()
recorder.add_figure(f, "first plot", svg=True)
plt.close(f)

# and save multiple figures at once
f1 = plt.figure()
f2 = plt.figure()

f1._save_name = "first of two"
f2._save_name = "second of two"
recorder.add_figures()

# finally, you can check what's saved so far
recorder.describe()



from setuptools import setup, find_namespace_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "numpy",
    "matplotlib",
    "fcutils",
    "myterial",
    "pyinspect",
    "rich",
    "loguru",
    "scipy",
]

setup(
    name="tpd",
    version="1.6",
    description="Easy saving of logs, data and plots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    install_requires=requirements,
    extras_require={},
    python_requires=">=3.6",
    packages=find_namespace_packages(exclude=("tests, examples")),
    entry_points={"console_scripts": []},
    include_package_data=True,
    url="https://github.com/FedeClaudi/dpl-store",
    author="Federico Claudi",
    zip_safe=False,
)



from pathlib import Path
from typing import Union
import numpy as np
from loguru import logger
from rich.logging import RichHandler
import matplotlib.pyplot as plt
from shutil import copyfile
from rich import print
from scipy.io import savemat
import json
import pandas as pd
import csv

from fcutils import path as fcpath
from fcutils.plot.figure import save_figure
from pyinspect.panels import Report
from myterial import salmon, green, green_light, blue_light, orange

from tpd import utils


def raise_if_not_started(fn):
    def wrapper(self, *args, **kwargs):
        if not self.started_status:
            raise ValueError(
                "Recorded must be started first, call `recorder.start()`"
            )
        return fn(self, *args, **kwargs)

    return wrapper


class Recorder:
    def __init__(self):
        self.started_status = False
        self.saved_figures = []
        self.saved_data = []
        self.folder = "not yet startd: call `recorder.start()...)`"
        self.name = None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Data Plots Logs - recorder: {self.folder}"

    @property
    def n_figures(self):
        """
            Number of saved figures
        """
        return len(self.saved_figures)

    @property
    def n_data(self):
        """
            Number of saved data fies
        """
        return len(self.saved_data)

    @property
    def n_files(self):
        """
            Total number of files in folder
        """
        return len(fcpath.files(self.folder))

    def start(
        self,
        base_folder: Union[str, Path] = None,
        name: str = None,
        timestamp: bool = True,
    ):
        self.started = utils.timestamp()
        self.started_status = True

        # get inputs
        self.base_folder = (
            Path(base_folder) if base_folder is not None else Path("./cache")
        )
        self.name = name or f"dpl_log"

        if timestamp:
            self.name += f"_{utils.timestamp()}"
        # create folders
        if not self.base_folder.exists():
            logger.warning(
                f'The base folder does not exist: "{self.base_folder}", creating it.'
            )
            self.base_folder.mkdir()

        self.folder = self.base_folder / self.name
        if not self.folder.exists():
            logger.warning(
                f'The destination folder does not exist: "{self.folder}", creating it.'
            )
            self.folder.mkdir(exist_ok=True)

        # start logging
        logger.configure(
            handlers=[
                {"sink": RichHandler(markup=True), "format": "{message}"}
            ]
        )
        log_file_path = self.folder / "log.log"
        if log_file_path.exists():
            log_file_path.unlink()
        logger.add(log_file_path)

        logger.debug(f"DPL - Saving data and logs to {self.folder}")
        logger.debug(f"Saving log file to: {str(log_file_path)}")

    @raise_if_not_started
    def copy(self, source: Union[str, Path]):
        """
            Copy a source file to the logs folder
        """
        source = Path(source)
        dest = self.folder / source.name

        logger.debug(f'Copying "{source}" to "{dest}"')
        if dest.exists():
            logger.warning("Source file exists already! OVERWRITING")
        copyfile(source, dest)

    @raise_if_not_started
    def add_data(
        self,
        data: np.ndarray,
        name: str,
        fmt: str = "npy",
        description: str = None,
    ):
        """
            Saves data from a numpy array to file.
            Can save to different formats (numpy, matlab...).
            If a 'description' is passed, a .txt file is saved describing what the
            data is. 
        """
        logger.debug(
            f'Saving data to file "{name}" with format {fmt}\nDescription: "{description}"'
        )

        # get destination path
        dest = self.folder / (name + "." + fmt)
        if dest.exists():
            logger.warning(
                f"Trying to save data to a file, but the destination file exists already - OVERWRITING"
            )

        # save data
        if fmt == "npy":
            np.save(dest, data)
        elif fmt == "mat":
            savemat(dest, mdict={name: data})
        elif fmt == "h5":
            # assumes the data are passed as a dataframe or dict
            utils.as_pandas(data).to_hdf(dest, key="hdf")
        elif fmt == "json":
            # assumes the data are passed as a dict
            with open(dest, "w", encoding="utf-8") as f:
                json.dump(data, f, ensure_ascii=False, indent=4)
        elif fmt == "csv":
            # assumes that data are a dict or pandas dataframe
            if isinstance(data, pd.DataFrame):
                data.to_csv(dest)
            else:
                with open(dest, "w", encoding="utf8", newline="") as f:
                    writer = csv.DictWriter(f, data.keys())
                    writer.writeheader()
                    writer.writerow(data)
        else:
            raise ValueError(f'Cannot save data to format: "{fmt}"')

        # save data description
        if description is not None:
            description_dest = self.folder / (name + ".txt")
            with open(description_dest, "w+") as dout:
                dout.write(description)

        self.saved_data.append(dest)

    @raise_if_not_started
    def add_figure(
        self, figure: Union[plt.Figure, plt.Axes], name: str, svg: bool = True
    ):
        """
            Saves a matplotlib figure to file
        """
        dest = self.folder / name
        logger.debug(f"Saving figure to: {dest}")

        if isinstance(figure, plt.Axes):
            figure = figure.figure
        save_figure(figure, dest, svg=svg)
        self.saved_figures.append(dest)

    @raise_if_not_started
    def add_figures(self, svg: bool = True):
        """
            Saves all open matplotlb figures, it assumes that each figure
            has a '._save_name' attribute with the name to use for saving.
        """
        # get all open figures
        figures = list(map(plt.figure, plt.get_fignums()))
        for figure in figures:
            try:
                self.add_figure(figure, figure._save_name, svg=svg)
            except AttributeError:
                logger.warning(
                    "Could not save an open figure since it did not have an `_save_name` attribute"
                )

    @raise_if_not_started
    def describe(self):
        if self.name is None:
            logger.warning(
                "recorder is not properly startd yet, call `recorder.start(...)` first"
            )
            return
        print("\n\n")
        rep = Report(
            title=f"LOGS: {self.name}",
            color=salmon,
            dim=orange,
            accent=orange,
            show_info=True,
        )

        rep.add(f'Logs saved at: [{blue_light}]"{self.folder}"')
        rep.add(f"Started at: [{blue_light}]{self.started}")

        # saved figures
        rep.spacer()
        rep.add(f"Saved: [bold {orange}]{len(self.saved_figures)}[/] figures:")
        figs_table = utils._make_table("#", "name")
        for n, fl in enumerate(self.saved_figures):
            figs_table.add_row(str(n), f"{fl.name}.png")

        rep.add(figs_table, "rich")

        # saved data
        rep.spacer()
        rep.add(f"Saved: [bold {orange}]{len(self.saved_data)}[/] data:")
        data_table = utils._make_table("#", "name")
        for n, fl in enumerate(self.saved_data):
            data_table.add_row(str(n), f"{fl.name}")
        rep.add(data_table, "rich")

        rep.print()

        # list files in destination folder
        files_table = utils._make_table("name", "size", nodim=True)
        for n, f in enumerate(fcpath.files(self.folder)):
            files_table.add_row(
                f"[{green_light}]{n} " + str(f.name),
                f"[{blue_light}]{fcpath.size(f)}",
            )
        print(f"\n[bold {green}]Files in destination folder ({self.folder})\n")
        print(files_table)
        print(
            f"[dim {green_light}]Total number of files: [dim {blue_light}]{self.n_files}"
        )


if __name__ == "__main__":
    rec = Recorder()
    rec.start()
    rec.add_data([1, 2], "test")
